// ----------------------------------------------------------------------
// SetPerms.cpp - set POSIX permission bits on selected files (Tracker)
//
// By Chris Herborth (chrish@qnx.com)
// Jan. 15, 1998
//
// Nicer UI by Marco Nelissen (marcone@xs4all.nl)
// March, 1998
// [I would've gotten around to tweaking the UI.  Marco beat me to it! - cjh]
//
// Updated for R4.
// Better exit using an atomic variable and a semaphore, suggested by
// Jon Watte.
// December, 1998

// ----------------------------------------------------------------------
// Standard headers
#include <storage/Entry.h>
#include <storage/Node.h>
#include <storage/File.h>
#include <storage/Directory.h>
#include <storage/SymLink.h>
#include <interface/Alert.h>
#include <interface/Point.h>
#include <interface/ScrollBar.h>
#include <kernel/OS.h>

#include <app/Application.h>

#include <new.h>

#include <stdio.h>
#include <stdlib.h>

// ----------------------------------------------------------------------
// Local headers
#include "MWindow.h"		// liblayout window
#include "layout.h"			// liblayout view
#include "HGroup.h"
#include "VGroup.h"
#include "MCheckBox.h"
#include "MButton.h"
#include "MStringView.h"
#include "MSplitter.h"
#include "Space.h"
#include "MBorder.h"

// ----------------------------------------------------------------------
// Exported functions
#include <add-ons/tracker/TrackerAddOn.h>

// ----------------------------------------------------------------------
// Local prototypes
void DoError( const char *message );

// ----------------------------------------------------------------------
// Constants
static const char *WINDOW_TITLE = "SetPerms 5";

enum { MSG_SET_USER_R  = 'u_r_',
	   MSG_SET_USER_W  = 'u_w_',
	   MSG_SET_USER_X  = 'u_x_',
	   MSG_SET_GRP_R   = 'g_r_',
	   MSG_SET_GRP_W   = 'g_w_',
	   MSG_SET_GRP_X   = 'g_x_',
	   MSG_SET_OTHER_R = 'o_r_',
	   MSG_SET_OTHER_W = 'o_w_',
	   MSG_SET_OTHER_X = 'o_x_',
	   MSG_CANCEL      = '-chg',
	   MSG_SET         = '+chg'
};


// ----------------------------------------------------------------------
// PermsWindow - the UI for editing a file's permission bits.
class PermsWindow : public MWindow
{
public:
	PermsWindow( BNode *node, entry_ref &ref, sem_id s, volatile int32 *c );
	~PermsWindow( void );
	void BuildWindow( void );

	virtual void MessageReceived( BMessage *msg );

	bool IsValid( void ) { return _valid; };

	status_t SetPermissions( void );
	
	int32 GetCounter( void ) { return (int)_window_counter; }
	
	void IncrCounter( void ) {
		(void)atomic_add( (int32 *)_window_counter, 1 );
	}
	
	void DecrCounter( void ) {
		(void)atomic_add( (int32 *)_window_counter, -1 );
	}

private:
	volatile int32 *_window_counter;
	sem_id _sem;

	BNode *_node;
	char _name[B_FILE_NAME_LENGTH];
	bool _valid;
	mode_t _mode;
	mode_t _new_mode;

	MView *_topview;
	
	static BRect _offset;
};

// Make the windows tiny; they'll resize automagically once we've
// added some views to them.
BRect PermsWindow::_offset( 50, 50, 60, 60 );

PermsWindow::PermsWindow( BNode *node, entry_ref &ref, sem_id sem, volatile int32 *counter )
	: MWindow( _offset, WINDOW_TITLE, B_TITLED_WINDOW, 
			   B_WILL_ACCEPT_FIRST_CLICK ),
	  _node( node ),
	  _valid( false ),
	  _sem( sem ),
	  _window_counter( counter )
{
	IncrCounter();

	// Now move our _offset.
	_offset.OffsetBy( B_V_SCROLL_BAR_WIDTH, B_H_SCROLL_BAR_HEIGHT );

	// First, try to find the file/folder name and permission bits.
	BEntry entry( &ref );
	if( entry.InitCheck() != B_NO_ERROR ) {
		// TODO: BAlert with an error
		DoError( "Can't create entry from node." );
		return;
	}

	if( entry.GetName( _name ) != B_NO_ERROR ) {
		// TODO: BAlert with an error
		DoError( "Can't get name from entry." );
		return;
	}

	if( node->GetPermissions( &_mode ) != B_NO_ERROR ) {
		// TODO: BAlert with an error
		DoError( "Can't get permissions from node." );
		return;
	}
	_new_mode = _mode;

	// If we get this far, everything is working.
	_valid = true;
}

PermsWindow::~PermsWindow( void )
{
	DecrCounter();
	release_sem( _sem );
	
	if( GetCounter() == 0 ) {
		_offset.OffsetTo( 50, 50 );
	}
}

void PermsWindow::BuildWindow( void )
{
	// Now we can build the window's contents.
	SetTitle(_name);

	_topview = 
		new MBorder
		(
			M_RAISED_BORDER,3,"",
			new VGroup
			(
				new HGroup
				(
					new MBorder
					(
						M_LABELED_BORDER,6,"Owner",
						new VGroup
						(
							new MCheckBox( "read", new BMessage( MSG_SET_USER_R ), this,  _mode & S_IRUSR ),
							new MCheckBox( "write", new BMessage( MSG_SET_USER_W ), this, _mode & S_IWUSR ),
							new MCheckBox( "execute", new BMessage( MSG_SET_USER_X ), this, _mode & S_IXUSR ),
							0
						)
					),
					new MBorder
					(
						M_LABELED_BORDER,6,"Group",
						new VGroup
						(
							new MCheckBox( "read",  new BMessage( MSG_SET_GRP_R ), this,  _mode & S_IRGRP ),
							new MCheckBox( "write",  new BMessage( MSG_SET_GRP_W ), this,  _mode & S_IWGRP ),
							new MCheckBox( "execute", new BMessage( MSG_SET_GRP_X ), this, _mode & S_IXGRP ),
							0
						)
					),
					new MBorder
					(
						M_LABELED_BORDER,6,"Other",
						new VGroup
						(
							new MCheckBox( "read", new BMessage( MSG_SET_OTHER_R ), this,  _mode & S_IROTH ),
							new MCheckBox( "write", new BMessage( MSG_SET_OTHER_W ), this, _mode & S_IWOTH ),
							new MCheckBox( "execute",new BMessage( MSG_SET_OTHER_X ), this, _mode & S_IXOTH ),
							0
						)
					),
					0
				),
			new HGroup
			(
				new MButton( "Cancel", new BMessage( MSG_CANCEL ), this, minimax(-1,-1,-1,-1) ),
				new MButton( "Set", new BMessage( MSG_SET ), this, minimax(-1,-1,-1,-1) ),
				0
			),
			0
		)
	);

	BView *be_topview = dynamic_cast<BView *>( _topview );

	AddChild( be_topview );
}

status_t PermsWindow::SetPermissions( void )
{
	if( _mode != _new_mode ) {
		return _node->SetPermissions( _new_mode );
	}

	return B_NO_ERROR;
}

void PermsWindow::MessageReceived( BMessage *msg )
{
	status_t retval;

	switch( msg->what ) {
	case MSG_SET_USER_R:
		_new_mode ^= S_IRUSR;
		break;

	case MSG_SET_USER_W:
		_new_mode ^= S_IWUSR;
		break;

	case MSG_SET_USER_X:
		_new_mode ^= S_IXUSR;
		break;

	case MSG_SET_GRP_R:
		_new_mode ^= S_IRGRP;
		break;

	case MSG_SET_GRP_W:
		_new_mode ^= S_IWGRP;
		break;

	case MSG_SET_GRP_X:
		_new_mode ^= S_IXGRP;
		break;

	case MSG_SET_OTHER_R:
		_new_mode ^= S_IROTH;
		break;

	case MSG_SET_OTHER_W:
		_new_mode ^= S_IWOTH;
		break;

	case MSG_SET_OTHER_X:
		_new_mode ^= S_IXOTH;
		break;

	case MSG_CANCEL:
		if( _mode != _new_mode ) {
			// TODO: BAlert asking if you should set it anyway
			DoError( "Should ask if you want to set the permissions "
					 "anyway..." );
			;
		}
		Quit();
		break;

	case MSG_SET:
		retval = SetPermissions();
		if( retval == B_NOT_ALLOWED ) {
			// TODO: BAlert error - can't set permissions on read-only volume
			DoError( "Can't set permissions on a read-only Bvolume." );
		} else if( retval == B_BAD_VALUE ) {
			// TODO: BAlert error - node doesn't exist
			DoError( "That node doesn't exist anymore." );
		} else if( retval != B_NO_ERROR ) {
			// TODO: BAlert error - unknown error!
			DoError( "Unknown error!" );
		}
		Quit();
		break;

	default:
		MWindow::MessageReceived( msg );
	}

	// TODO: Do _I_ have to delete msg?
}

// ----------------------------------------------------------------------
// Gimpy error handler.
void DoError( const char *message )
{
	try {
		BAlert *alert = new BAlert( "SetPerms Error",
									message,
									"Dang", NULL, NULL,
									B_WIDTH_AS_USUAL,
									B_STOP_ALERT );
		int32 retval = alert->Go();
	} catch ( bad_alloc &ex ) {
		printf( "aieee... - %s\n", message );
	}
}

// ----------------------------------------------------------------------
// Handle the files/folders selected in the Tracker window.
void process_refs( entry_ref dir_ref, BMessage *msg, void *data )
{
#pragma unused (dir_ref,data)
	long count = 0;
	uint32 type = 0;
	
	sem_id sem = create_sem( 0, "SetPerms exit" );
	if( sem < B_OK ) {
		// That's bad, let's bail.
		// TODO: should warn somebody...
		return;
	}
	volatile int32 counter = 0;

	msg->GetInfo( "refs", &type, &count );
	if( count == 0 ) {
		// TODO: BAlert error - select files/folders first
		DoError( "Select files/folders first!" );
		return;
	}

	for( uint32 idx = 0; idx < count; idx++ ) {
		// Get the entry_ref, then fire up a window for it.
		entry_ref the_ref;

		status_t retval = msg->FindRef( "refs", idx, &the_ref );
		if( retval != B_NO_ERROR ) {
			// TODO: BAlert error - can't read refs!
			DoError( "Can't read refs from message!" );
		}

		// Hmm, we should probably throttle this so we don't eat all the fds...
		BNode *the_node = new BNode( &the_ref );
		if( the_node->InitCheck() != B_NO_ERROR ) {
			// TODO: BAlert error - whoops, too many nodes
			DoError( "Woah, I've opened too many nodes..." );
		} else {
			try {
				PermsWindow *win = new PermsWindow( the_node, the_ref, sem, &counter );
				win->BuildWindow();
				win->Show();
			} catch ( bad_alloc &ex ) {
				DoError( "Can't create PermsWindow!" );
			}
		}
	}
	
	// For a Tracker add-on, we can't leave until all of our windows are done.
	do {
		acquire_sem( sem );
	} while( counter > 0 );
}

// ----------------------------------------------------------------------
// main() and BApplication to make this a standalone program as well
// as a Tracker add-on.

class SetPermsApplication : public BApplication {
public:
	SetPermsApplication( void )
		: BApplication( "application/x-vnd.ads.SetPerms" ) { ; }
	virtual ~SetPermsApplication( void ) { ; }

	virtual void RefsReceived( BMessage *msg ) {
		entry_ref ref;
		BEntry ent( "/boot" );
		(void)ent.GetRef( &ref );
		process_refs( ref, msg, NULL );
	}
};

int main( void )
{
	SetPermsApplication app;
	app.PostMessage( B_QUIT_REQUESTED );
	app.Run();
	return EXIT_SUCCESS;
}
